import 'dart:io';

import 'package:camera/camera.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:google_mlkit_barcode_scanning/google_mlkit_barcode_scanning.dart';
import 'package:vibration/vibration.dart';

class BarcodeScannerView extends StatefulWidget {
  const BarcodeScannerView({
      Key? key,
      required this.onDetected,
      this.detectionInterval = 1000,
      this.initialCameraLensDirection = CameraLensDirection.back,
      this.overlayWidgetBuilder})
    : super(key: key);

  final Function(List<Barcode> barcodes) onDetected;

  /// The interval in milliseconds at which to process frames for detection
  final int detectionInterval;

  final CameraLensDirection initialCameraLensDirection;
  final Widget? Function(List<Barcode>)? overlayWidgetBuilder;


  @override
  State<BarcodeScannerView> createState() => _BarcodeScannerViewState();
}

class _BarcodeScannerViewState extends State<BarcodeScannerView> {
  final BarcodeScanner _barcodeScanner = BarcodeScanner();
  bool _canProcess = true;
  bool _isBusy = false;
  String? _text;

  var _cameraLensDirection = CameraLensDirection.back;

  static List<CameraDescription> _cameras = [];
  CameraController? _controller;
  int _cameraIndex = -1;
  double _currentZoomLevel = 1.0;
  double _minAvailableZoom = 1.0;
  double _maxAvailableZoom = 1.0;
  double _minAvailableExposureOffset = 0.0;
  double _maxAvailableExposureOffset = 0.0;
  double _currentExposureOffset = 0.0;
  bool _changingCameraLens = false;

  int scanningMode = 0;
  final List<bool> _selectedScanningModes = <bool>[true, false, false];

  static const double _activeScanBtnSize = 74;
  final double _activeScanBtnCompliment = _activeScanBtnSize * 0.66;

  Widget? overlayWidget = null;

  int lastProcessTS = 0;


  @override
  void initState() {
    super.initState();

    _initialize();
  }

  void _initialize() async {
    if (_cameras.isEmpty) {
      _cameras = await availableCameras();
    }
    for (var i = 0; i < _cameras.length; i++) {
      if (_cameras[i].lensDirection == widget.initialCameraLensDirection) {
        _cameraIndex = i;
        break;
      }
    }
    if (_cameraIndex != -1) {
      _startLiveFeed();
    }
  }

  @override
  void dispose() {
    _canProcess = false;
    _barcodeScanner.close();
    super.dispose();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    final CameraController? cameraController = _controller;

    if (state == AppLifecycleState.inactive) {
      _stopLiveFeed();
    } else if (state == AppLifecycleState.resumed) {
      _startLiveFeed();
    }
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      extendBodyBehindAppBar: true,
      extendBody: true,
      appBar: AppBar(
        backgroundColor: Colors.transparent,
        elevation: 0,
        actions: [
          IconButton(
            icon: (_controller?.value.flashMode ?? FlashMode.off) != FlashMode.torch
                ? const Icon(Icons.flash_off, color: Colors.white)
                : const Icon(Icons.flash_on, color: Colors.yellow),
            iconSize: 32.0,
            onPressed: _controller != null
                ? () => _controller!.setFlashMode(_controller!.value.flashMode == FlashMode.torch ? FlashMode.off : FlashMode.torch)
                  .onError((error, stackTrace) {
                    debugPrint("Failed to toggle FlashMode");
                    debugPrintStack(stackTrace: stackTrace);
                  })
                  .then((_) {
                    if (mounted) {
                      setState(() {});
                    }
                  })
                : null,
          ),
        ],
      ),
      body: _cameraWidget(),
      bottomNavigationBar: BottomAppBar(
        padding: const EdgeInsets.only(left: 8, right: 8, top: 8),
        color: const Color.fromRGBO(0, 0, 0, 0.6),
        surfaceTintColor: Colors.black,
        height: 128,
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          children: [
            Container(
              constraints: const BoxConstraints.expand(width: double.infinity, height: 22),
              child: ToggleButtons(
                direction: Axis.horizontal,
                onPressed: (int index) {
                  setState(() {
                    for (int i = 0; i < _selectedScanningModes.length; i++) {
                      _selectedScanningModes[i] = i == index;
                    }
                    scanningMode = index;
                    _canProcess = index == 0;
                  });
                },
                selectedBorderColor: Colors.transparent,
                selectedColor: Colors.white,
                fillColor: Colors.transparent,
                borderColor: Colors.transparent,
                splashColor: Colors.transparent,
                renderBorder: false,
                color: Colors.white30,
                constraints: BoxConstraints.expand(height: 22, width: (MediaQuery.of(context).size.width - 16) /3,),
                isSelected: _selectedScanningModes,
                children: const [
                  Text("CONTINUOUS"),
                  Text("PRESS & HOLD"),
                  Text("TOGGLE")
                ],
              ),
            ),
            SizedBox.fromSize(size: const Size(1, 8)),
            Row(
              children: [
                Container(width: 64),

                Expanded(child: Container()),

                Stack(
                  alignment: Alignment.center,
                  children: [
                    Container(
                      height: 90,
                      width: 90,
                      decoration: BoxDecoration(
                          border: Border.all(color: Colors.white, width: 4),
                          color: Colors.transparent,
                          shape: BoxShape.circle
                      ),
                    ),

                    GestureDetector(
                      onTapDown: (tap) {
                        setState(() {
                          if (scanningMode != 0) {
                            _canProcess = !_canProcess;
                          }
                        });
                      },
                      onTapUp: (tap) {
                        if (scanningMode == 1) {
                          setState(() {
                            _canProcess = !_canProcess;
                          });
                        }
                      },
                      child: Visibility(
                        visible: !_canProcess,
                        replacement: Container(
                            constraints: BoxConstraints.tight(Size(_activeScanBtnCompliment, _activeScanBtnCompliment)),
                            decoration: BoxDecoration(
                              border: Border.all(color:  const Color.fromRGBO(231, 76, 70, 1), width: 0),
                              shape: BoxShape.rectangle,
                              borderRadius: BorderRadius.all(Radius.circular(_activeScanBtnCompliment /4)),
                              color: const Color.fromRGBO(231, 76, 70, 1),
                            )
                        ),
                        child: Container(
                          height: _activeScanBtnSize,
                          width: _activeScanBtnSize,
                          decoration: BoxDecoration(
                              border: Border.all(color: Colors.white, width: 0),
                              shape: BoxShape.circle,
                              color: Colors.white
                          ),
                        ),
                      ),
                    ),
                  ],
                ),
                Expanded(child: Container()),

                RawMaterialButton(
                  constraints: BoxConstraints.tight(const Size(56,56)),
                  onPressed: () {
                    setState(() {
                      _canProcess = false;
                    });
                  },
                  elevation: 2.0,
                  fillColor: const Color.fromRGBO(44, 44, 44, 1),
                  padding: const EdgeInsets.all(0.0),
                  shape: const CircleBorder(),
                  child: IconButton(
                    icon: const Icon(Icons.flip_camera_android, color: Colors.white),
                    constraints: BoxConstraints.tight(const Size(56,56)),
                    onPressed: () {
                      _switchLiveCamera();
                    },
                  ),
                ),
              ],
            )
          ],
        ),
      ),
    );
  }

  Widget _cameraWidget() {
    if (_cameras.isEmpty) return Container();
    if (_controller == null) return Container();
    if (_controller?.value.isInitialized == false) return Container();
    final size = MediaQuery.of(context).size;
    final deviceRatio = size.width / size.height;
    final double xScale = _controller !=  null ? _controller!.value.aspectRatio / deviceRatio : 1;

    return SizedBox(
      width: size.width,
      height: size.height,
      child: Container(
        color: Colors.black,
        child: Stack(
          fit: StackFit.expand,
          children: <Widget>[
            Center(
              child: _changingCameraLens
                ? const Center(
                    child: Text('Changing camera lens'),
                  )
                : Transform.scale(
                  scale: 1 / (_controller!.value.aspectRatio * size.aspectRatio),
                  alignment: Alignment.center,
                  child: CameraPreview(
                      _controller!,
                      child: overlayWidget,
                    ),
                )
              )
          ],
        ),
      ),
    );
  }

  Future _startLiveFeed() async {
    final camera = _cameras[_cameraIndex];
    _controller = CameraController(
      camera,
      // Set to ResolutionPreset.high. Do NOT set it to ResolutionPreset.max because for some phones does NOT work.
      ResolutionPreset.high,
      enableAudio: false,
      imageFormatGroup: Platform.isAndroid
          ? ImageFormatGroup.nv21
          : ImageFormatGroup.bgra8888,
    );
    _controller?.initialize().then((_) {
      if (!mounted) {
        return;
      }
      _initLiveFeed();
    });
  }

  Future _stopLiveFeed() async {
    await _controller?.stopImageStream();
    await _controller?.dispose();
    _controller = null;
  }

  void _initLiveFeed() {
    _controller?.getMinZoomLevel().then((value) {
      _currentZoomLevel = value;
      _minAvailableZoom = value;
    });
    _controller?.getMaxZoomLevel().then((value) {
      _maxAvailableZoom = value;
    });
    _currentExposureOffset = 0.0;
    _controller?.getMinExposureOffset().then((value) {
      _minAvailableExposureOffset = value;
    });
    _controller?.getMaxExposureOffset().then((value) {
      _maxAvailableExposureOffset = value;
    });
    _controller?.startImageStream(_processCameraImage);

    _controller?.lockCaptureOrientation(DeviceOrientation.portraitUp);

    setState(() {});
  }

  Future _switchLiveCamera() async {
    setState(() => _changingCameraLens = true);
    _cameraIndex = (_cameraIndex + 1) % _cameras.length;

    //await _stopLiveFeed();
    //await _startLiveFeed();
    if (_controller != null) {
      await _controller!.stopImageStream();
      await _controller!.setDescription(_cameras[_cameraIndex]);
      _initLiveFeed();
    } else {
      await _startLiveFeed();
    }

    setState(() => _changingCameraLens = false);
  }

  final _orientations = {
    DeviceOrientation.portraitUp: 0,
    DeviceOrientation.landscapeLeft: 90,
    DeviceOrientation.portraitDown: 180,
    DeviceOrientation.landscapeRight: 270,
  };


  void _processCameraImage(CameraImage image) {
    if ((lastProcessTS - DateTime.now().millisecondsSinceEpoch).abs() < widget.detectionInterval) {
      return;
    }

    final inputImage = _inputImageFromCameraImage(image);
    if (inputImage == null) return;

    _processImage(inputImage);
    lastProcessTS = DateTime.now().millisecondsSinceEpoch;
  }

  InputImage? _inputImageFromCameraImage(CameraImage image) {
    if (_controller == null) return null;

    // get image rotation
    // it is used in android to convert the InputImage from Dart to Java: https://github.com/flutter-ml/google_ml_kit_flutter/blob/master/packages/google_mlkit_commons/android/src/main/java/com/google_mlkit_commons/InputImageConverter.java
    // `rotation` is not used in iOS to convert the InputImage from Dart to Obj-C: https://github.com/flutter-ml/google_ml_kit_flutter/blob/master/packages/google_mlkit_commons/ios/Classes/MLKVisionImage%2BFlutterPlugin.m
    // in both platforms `rotation` and `camera.lensDirection` can be used to compensate `x` and `y` coordinates on a canvas: https://github.com/flutter-ml/google_ml_kit_flutter/blob/master/packages/example/lib/vision_detector_views/painters/coordinates_translator.dart
    final camera = _cameras[_cameraIndex];
    final sensorOrientation = camera.sensorOrientation;
    // print(
    //     'lensDirection: ${camera.lensDirection}, sensorOrientation: $sensorOrientation, ${_controller?.value.deviceOrientation} ${_controller?.value.lockedCaptureOrientation} ${_controller?.value.isCaptureOrientationLocked}');
    InputImageRotation? rotation;
    if (Platform.isIOS) {
      rotation = InputImageRotationValue.fromRawValue(sensorOrientation);
    } else if (Platform.isAndroid) {
      var rotationCompensation =
      _orientations[_controller!.value.deviceOrientation];
      if (rotationCompensation == null) return null;
      if (camera.lensDirection == CameraLensDirection.front) {
        // front-facing
        rotationCompensation = (sensorOrientation + rotationCompensation) % 360;
      } else {
        // back-facing
        rotationCompensation =
            (sensorOrientation - rotationCompensation + 360) % 360;
      }
      rotation = InputImageRotationValue.fromRawValue(rotationCompensation);
      // print('rotationCompensation: $rotationCompensation');
    }
    if (rotation == null) return null;
    // print('final rotation: $rotation');

    // get image format
    final format = InputImageFormatValue.fromRawValue(image.format.raw);
    // validate format depending on platform
    // only supported formats:
    // * nv21 for Android
    // * bgra8888 for iOS
    if (format == null ||
        (Platform.isAndroid && format != InputImageFormat.nv21) ||
        (Platform.isIOS && format != InputImageFormat.bgra8888)) return null;

    // since format is constraint to nv21 or bgra8888, both only have one plane
    if (image.planes.length != 1) return null;
    final plane = image.planes.first;

    // compose InputImage using bytes
    return InputImage.fromBytes(
      bytes: plane.bytes,
      metadata: InputImageMetadata(
        size: Size(image.width.toDouble(), image.height.toDouble()),
        rotation: rotation, // used only in Android
        format: format, // used only in iOS
        bytesPerRow: plane.bytesPerRow, // used only in iOS
      ),
    );
  }

  Future<void> _processImage(InputImage inputImage) async {
    if (!_canProcess) return;
    if (_isBusy) return;
    _isBusy = true;
    setState(() {
      _text = '';
    });
    final barcodes = await _barcodeScanner.processImage(inputImage);

    if (barcodes.isNotEmpty) {
      if (Platform.isIOS) {
        HapticFeedback.vibrate();
      } else if (Platform.isAndroid) {
        if (await Vibration.hasVibrator() ?? false) {
          Vibration.vibrate();
        }
      }
      widget.onDetected(barcodes);
    }
    _isBusy = false;
    if (mounted) {
      setState(() {
        if (widget.overlayWidgetBuilder != null) {
          overlayWidget = widget.overlayWidgetBuilder!(barcodes);
        }
      });
    }
  }
}
